import hashlib
import os
import random
import tempfile
from datetime import timedelta
from unittest.mock import Mock

from django.utils import timezone

import pytest
from vintasend.constants import NotificationStatus, NotificationTypes
from vintasend.exceptions import (
    NotificationCancelError,
    NotificationNotFoundError,
    NotificationUpdateError,
)
from vintasend.services.dataclasses import Notification, NotificationAttachment, OneOffNotification

from vintasend_django.models import Attachment as AttachmentModel
from vintasend_django.models import Notification as NotificationModel
from vintasend_django.services.notification_backends.django_db_notification_backend import (
    DjangoDbNotificationBackend,
)
from vintasend_django.test_helpers import VintaSendDjangoTestCase


class DjangoDBNotificationBackendTestCase(VintaSendDjangoTestCase):
    def setUp(self):
        super().setUp()
        self.temp_files = []  # Track temporary files created during tests

    def tearDown(self):
        super().tearDown()
        # Clean up temporary files
        for file_path in self.temp_files:
            try:
                if os.path.exists(file_path):
                    os.unlink(file_path)
            except OSError:
                pass  # File might already be deleted

        # Clean up attachment files from Django's file storage
        # This will clean up files created by attachment tests
        for attachment in AttachmentModel.objects.all():
            try:
                if attachment.file:
                    attachment.file.delete(save=False)
            except OSError:
                pass  # File might already be deleted

        # Additional cleanup: manually remove any remaining files in the attachments directory
        import glob
        attachment_dir = "notifications/attachments"
        if os.path.exists(attachment_dir):
            # Remove all files in the attachments directory
            for file_path in glob.glob(os.path.join(attachment_dir, "*")):
                try:
                    if os.path.isfile(file_path):
                        os.unlink(file_path)
                except OSError:
                    pass  # File might already be deleted

    def test_persist_notification(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        assert isinstance(notification, Notification)
        assert str(notification.user_id) == str(self.user.pk)
        assert notification.notification_type == NotificationTypes.EMAIL.value
        assert notification.title == "test"
        assert notification.body_template == "test"
        assert notification.context_name == "test"
        assert notification.context_kwargs == {}
        assert notification.send_after is None
        assert notification.subject_template == "test"
        assert notification.preheader_template == "test"
        assert notification.status == NotificationStatus.PENDING_SEND.value
        assert notification.id is not None
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert str(notification_db_record.user_id) == str(self.user.pk)
        assert notification_db_record.notification_type == NotificationTypes.EMAIL.value
        assert notification_db_record.title == "test"
        assert notification_db_record.body_template == "test"
        assert notification_db_record.context_name == "test"
        assert notification_db_record.context_kwargs == {}
        assert notification_db_record.send_after is None
        assert notification_db_record.subject_template == "test"
        assert notification_db_record.preheader_template == "test"
        assert notification_db_record.status == NotificationStatus.PENDING_SEND.value

    def test_update_notification(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        updated_notification = DjangoDbNotificationBackend().persist_notification_update(
            notification_id=notification.id,
            updated_data={"subject_template": "updated test subject"},
        )

        assert updated_notification.subject_template == "updated test subject"
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.subject_template == "updated test subject"

    def get_all_pending_notifications(self):
        DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test 2",
            body_template="test 2",
            context_name="test 2",
            context_kwargs={},
            send_after=None,
            subject_template="test 2",
            preheader_template="test 2",
        )

        already_sent = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test already sent",
            body_template="test already sent",
            context_name="test already sent",
            context_kwargs={},
            send_after=None,
            subject_template="test already sent",
            preheader_template="test already sent",
        )
        DjangoDbNotificationBackend().mark_pending_as_sent(notification_id=already_sent.id)

        notifications = DjangoDbNotificationBackend().get_all_pending_notifications()
        assert len(notifications) == 2
        notification_1 = notifications[1]
        assert isinstance(notification_1, Notification)
        assert notification_1.user_id == self.user.pk
        assert notification_1.notification_type == NotificationTypes.EMAIL.value
        assert notification_1.title == "test"
        assert notification_1.body_template == "test"
        assert notification_1.context_name == "test"
        assert notification_1.context_kwargs == {}
        assert notification_1.send_after is None
        assert notification_1.subject_template == "test"
        assert notification_1.preheader_template == "test"
        assert notification_1.status == NotificationStatus.PENDING_SEND.value
        notification_2 = notifications[0]
        assert isinstance(notification_2, Notification)
        assert notification_2.user_id == self.user.pk
        assert notification_2.notification_type == NotificationTypes.EMAIL.value
        assert notification_2.title == "test 2"
        assert notification_2.body_template == "test 2"
        assert notification_2.context_name == "test 2"
        assert notification_2.context_kwargs == {}
        assert notification_2.send_after is None
        assert notification_2.subject_template == "test 2"
        assert notification_2.preheader_template == "test 2"
        assert notification_2.status == NotificationStatus.PENDING_SEND.value

    def test_get_pending_notifications(self):
        DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test 2",
            body_template="test 2",
            context_name="test 2",
            context_kwargs={},
            send_after=None,
            subject_template="test 2",
            preheader_template="test 2",
        )

        already_sent = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test already sent",
            body_template="test already sent",
            context_name="test already sent",
            context_kwargs={},
            send_after=None,
            subject_template="test already sent",
            preheader_template="test already sent",
        )
        DjangoDbNotificationBackend().persist_notification_update(
            notification_id=already_sent.id,
            updated_data={"status": NotificationStatus.SENT.value},
        )

        notifications = list(
            DjangoDbNotificationBackend().get_pending_notifications(page=1, page_size=1)
        )
        assert len(notifications) == 1
        notification_1 = notifications[0]
        assert isinstance(notification_1, Notification)
        assert notification_1.user_id == self.user.pk
        assert notification_1.notification_type == NotificationTypes.EMAIL.value
        assert notification_1.title == "test"
        assert notification_1.body_template == "test"
        assert notification_1.context_name == "test"
        assert notification_1.context_kwargs == {}
        assert notification_1.send_after is None
        assert notification_1.subject_template == "test"
        assert notification_1.preheader_template == "test"
        assert notification_1.status == NotificationStatus.PENDING_SEND.value

        notifications = list(
            DjangoDbNotificationBackend().get_pending_notifications(page=2, page_size=1)
        )
        assert len(notifications) == 1
        notification_2 = notifications[0]
        assert isinstance(notification_2, Notification)
        assert notification_2.user_id == self.user.pk
        assert notification_2.notification_type == NotificationTypes.EMAIL.value
        assert notification_2.title == "test 2"
        assert notification_2.body_template == "test 2"
        assert notification_2.context_name == "test 2"
        assert notification_2.context_kwargs == {}
        assert notification_2.send_after is None
        assert notification_2.subject_template == "test 2"
        assert notification_2.preheader_template == "test 2"
        assert notification_2.status == NotificationStatus.PENDING_SEND.value

    def test_mark_pending_as_sent(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        notification = DjangoDbNotificationBackend().mark_pending_as_sent(notification.id)
        assert notification.status == NotificationStatus.SENT.value
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status == NotificationStatus.SENT.value

    def test_mark_pending_as_failed(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        notification = DjangoDbNotificationBackend().mark_pending_as_failed(notification.id)
        assert notification.status == NotificationStatus.FAILED.value
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status == NotificationStatus.FAILED.value

    def test_mark_pending_as_failed_already_sent(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )
        DjangoDbNotificationBackend().mark_pending_as_sent(notification.id)

        with pytest.raises(NotificationUpdateError):
            DjangoDbNotificationBackend().mark_pending_as_failed(notification.id)
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status == NotificationStatus.SENT.value

    def test_mark_sent_as_read(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )
        DjangoDbNotificationBackend().mark_pending_as_sent(notification.id)

        notification = DjangoDbNotificationBackend().mark_sent_as_read(notification.id)
        assert notification.status == NotificationStatus.READ.value
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status == NotificationStatus.READ.value

    def test_cancel_notification(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=timezone.now() + timedelta(days=1),
            subject_template="test",
            preheader_template="test",
        )

        DjangoDbNotificationBackend().cancel_notification(notification.id)
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status == NotificationStatus.CANCELLED.value

    def test_cancel_notification_already_sent(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )
        DjangoDbNotificationBackend().mark_pending_as_sent(notification.id)

        with pytest.raises(NotificationCancelError):
            DjangoDbNotificationBackend().cancel_notification(notification.id)
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        assert notification_db_record.status != NotificationStatus.CANCELLED.value

    def test_get_notification(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )

        notification_retrieved = DjangoDbNotificationBackend().get_notification(notification.id)
        assert notification_retrieved.id == notification.id
        assert notification_retrieved.user_id == self.user.pk
        assert notification_retrieved.notification_type == NotificationTypes.EMAIL.value
        assert notification_retrieved.title == "test"
        assert notification_retrieved.body_template == "test"
        assert notification_retrieved.context_name == "test"
        assert notification_retrieved.context_kwargs == {}
        assert notification_retrieved.send_after is None
        assert notification_retrieved.subject_template == "test"
        assert notification_retrieved.preheader_template == "test"
        assert notification_retrieved.status == NotificationStatus.PENDING_SEND.value

    def test_get_notification_not_found(self):
        with pytest.raises(NotificationNotFoundError):
            DjangoDbNotificationBackend().get_notification(random.randint(1, 100))

    def test_get_notification_cancelled(self):
        notification = DjangoDbNotificationBackend().persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            subject_template="test",
            preheader_template="test",
        )
        DjangoDbNotificationBackend().cancel_notification(notification.id)
        with pytest.raises(NotificationNotFoundError):
            DjangoDbNotificationBackend().get_notification(notification.id)

    def test_persist_one_off_notification(self):
        """Test creating one-off notification"""
        backend = DjangoDbNotificationBackend()
        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="test@example.com",
            first_name="John",
            last_name="Doe",
            notification_type=NotificationTypes.EMAIL.value,
            title="Welcome Email",
            body_template="welcome_email",
            context_name="welcome_context",
            context_kwargs={"user_name": "John"},
            send_after=None,
            subject_template="Welcome to our platform",
            preheader_template="Get started today",
        )

        assert isinstance(one_off_notification, OneOffNotification)
        assert one_off_notification.email_or_phone == "test@example.com"
        assert one_off_notification.first_name == "John"
        assert one_off_notification.last_name == "Doe"
        assert one_off_notification.notification_type == NotificationTypes.EMAIL.value
        assert one_off_notification.title == "Welcome Email"
        assert one_off_notification.status == NotificationStatus.PENDING_SEND.value

        # Verify it's stored in database
        notification_db_record = NotificationModel.objects.get(id=one_off_notification.id)
        assert notification_db_record.email_or_phone == "test@example.com"
        assert notification_db_record.first_name == "John"
        assert notification_db_record.last_name == "Doe"
        assert notification_db_record.user is None  # One-off notifications have no user

    def test_get_one_off_notification(self):
        """Test retrieving one-off notification"""
        backend = DjangoDbNotificationBackend()

        # Create one-off notification
        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="test@example.com",
            first_name="Jane",
            last_name="Smith",
            notification_type=NotificationTypes.EMAIL.value,
            title="Test One-off",
            body_template="test_template",
            context_name="test_context",
            context_kwargs={},
        )

        # Retrieve via _get_one_off_notification method
        retrieved = backend._get_one_off_notification(one_off_notification.id)
        assert isinstance(retrieved, OneOffNotification)
        assert retrieved.id == one_off_notification.id
        assert retrieved.email_or_phone == "test@example.com"
        assert retrieved.first_name == "Jane"
        assert retrieved.last_name == "Smith"

    def test_get_notification_handles_both_types(self):
        """Test that get_notification works for both regular and one-off notifications"""
        backend = DjangoDbNotificationBackend()

        # Create regular notification
        regular_notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Regular notification",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
        )

        # Create one-off notification
        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="test@example.com",
            first_name="Test",
            last_name="User",
            notification_type=NotificationTypes.EMAIL.value,
            title="One-off notification",
            body_template="test",
            context_name="test",
            context_kwargs={},
        )

        # Retrieve both via get_notification
        retrieved_regular = backend.get_notification(regular_notification.id)
        retrieved_one_off = backend.get_notification(one_off_notification.id)

        assert isinstance(retrieved_regular, Notification)
        assert isinstance(retrieved_one_off, OneOffNotification)
        assert retrieved_regular.id == regular_notification.id
        assert retrieved_one_off.id == one_off_notification.id

    def test_get_all_pending_notifications_includes_one_off(self):
        """Test that get_all_pending_notifications returns both types"""
        backend = DjangoDbNotificationBackend()

        # Create regular notification
        regular_notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Regular notification",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
        )

        # Create one-off notification
        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="test@example.com",
            first_name="Test",
            last_name="User",
            notification_type=NotificationTypes.EMAIL.value,
            title="One-off notification",
            body_template="test",
            context_name="test",
            context_kwargs={},
        )

        # Get all pending notifications
        all_pending = list(backend.get_all_pending_notifications())

        # Should contain both notifications
        assert len(all_pending) >= 2

        # Check that we have both types
        regular_found = False
        one_off_found = False
        for notification in all_pending:
            if isinstance(notification, Notification) and notification.id == regular_notification.id:
                regular_found = True
            elif isinstance(notification, OneOffNotification) and notification.id == one_off_notification.id:
                one_off_found = True

        assert regular_found, "Regular notification not found in pending notifications"
        assert one_off_found, "One-off notification not found in pending notifications"

    def test_persist_notification_with_file_path_attachment(self):
        """Test persisting regular notification with file path attachment"""
        backend = DjangoDbNotificationBackend()

        # Create a temporary file for testing
        with tempfile.NamedTemporaryFile(mode='w', suffix='.txt', delete=False) as tmp_file:
            tmp_file.write("Test attachment content")
            tmp_file_path = tmp_file.name

        # Track temp file for cleanup
        self.temp_files.append(tmp_file_path)

        # Create mock attachment
        mock_attachment = Mock(spec=NotificationAttachment)
        mock_attachment.file_path = tmp_file_path

        notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test with attachment",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            attachments=[mock_attachment],
        )

        assert isinstance(notification, Notification)
        assert notification.user_id == str(self.user.pk)

        # Verify attachment was stored
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 1

        attachment = attachments[0]
        assert attachment.name == os.path.basename(tmp_file_path)
        assert attachment.mime_type == 'application/octet-stream'
        assert attachment.size > 0
        assert attachment.file  # File should be saved

    def test_persist_one_off_notification_with_file_path_attachment(self):
        """Test persisting one-off notification with file path attachment"""
        backend = DjangoDbNotificationBackend()

        # Create a temporary file for testing
        with tempfile.NamedTemporaryFile(mode='w', suffix='.txt', delete=False) as tmp_file:
            tmp_file.write("Test one-off attachment content")
            tmp_file_path = tmp_file.name

        # Track temp file for cleanup
        self.temp_files.append(tmp_file_path)

        # Create mock attachment
        mock_attachment = Mock(spec=NotificationAttachment)
        mock_attachment.file_path = tmp_file_path

        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="test@example.com",
            first_name="John",
            last_name="Doe",
            notification_type=NotificationTypes.EMAIL.value,
            title="One-off with attachment",
            body_template="test",
            context_name="test",
            context_kwargs={},
            attachments=[mock_attachment],
        )

        assert isinstance(one_off_notification, OneOffNotification)
        assert one_off_notification.email_or_phone == "test@example.com"

        # Verify attachment was stored
        notification_db_record = NotificationModel.objects.get(id=one_off_notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 1

        attachment = attachments[0]
        assert attachment.name == os.path.basename(tmp_file_path)
        assert attachment.mime_type == 'application/octet-stream'
        assert attachment.size > 0
        assert attachment.file  # File should be saved

    def test_store_attachments_with_file_path(self):
        """Test _store_attachments method with file path attachment"""
        backend = DjangoDbNotificationBackend()

        # Create a temporary file for testing
        with tempfile.NamedTemporaryFile(mode='w', suffix='.txt', delete=False) as tmp_file:
            tmp_file.write("Test content for attachment")
            tmp_file_path = tmp_file.name

        # Track the temp file for cleanup
        self.temp_files.append(tmp_file_path)

        # Create mock attachment with file_path
        mock_attachment = Mock(spec=NotificationAttachment)
        mock_attachment.file_path = tmp_file_path

        # Test the _store_attachments method directly
        stored_attachments = backend._store_attachments([mock_attachment])

        assert len(stored_attachments) == 1
        attachment = stored_attachments[0]

        # Verify attachment properties
        assert isinstance(attachment, AttachmentModel)
        assert attachment.name == os.path.basename(tmp_file_path)
        assert attachment.mime_type == 'application/octet-stream'
        assert attachment.size > 0
        assert attachment.file  # File should be saved but not yet linked to notification

        # Note: attachment.notification should be None since it's not yet linked
        assert attachment.notification_id is None

    def test_store_attachments_without_file_path(self):
        """Test _store_attachments method with attachment that doesn't have file_path"""
        backend = DjangoDbNotificationBackend()

        # Create mock attachment without file_path
        mock_attachment = Mock(spec=NotificationAttachment)
        # Deliberately not setting file_path attribute

        # Test the _store_attachments method directly
        with pytest.raises(ValueError):
            backend._store_attachments([mock_attachment])

    def test_store_attachments_multiple_files(self):
        """Test _store_attachments method with multiple file path attachments"""
        backend = DjangoDbNotificationBackend()

        # Create multiple temporary files for testing
        tmp_files = []
        for i in range(3):
            tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix=f'_{i}.txt', delete=False)
            tmp_file.write(f"Test content for attachment {i}")
            tmp_file.close()
            tmp_files.append(tmp_file.name)

        # Track temp files for cleanup
        self.temp_files.extend(tmp_files)

        # Create mock attachments
        mock_attachments = []
        for tmp_file_path in tmp_files:
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path
            mock_attachments.append(mock_attachment)

        # Test the _store_attachments method with multiple attachments
        stored_attachments = backend._store_attachments(mock_attachments)

        assert len(stored_attachments) == 3

        # Sort both lists by file name for consistent comparison
        stored_attachments_sorted = sorted(stored_attachments, key=lambda x: x.name)
        tmp_files_sorted = sorted(tmp_files, key=lambda x: os.path.basename(x))

        for i, attachment in enumerate(stored_attachments_sorted):
            assert isinstance(attachment, AttachmentModel)
            assert attachment.name == os.path.basename(tmp_files_sorted[i])
            assert attachment.mime_type == 'application/octet-stream'
            assert attachment.size > 0
            assert attachment.file

    def test_persist_notification_with_multiple_attachments(self):
        """Test persisting regular notification with multiple attachments"""
        backend = DjangoDbNotificationBackend()

        # Create multiple temporary files for testing
        tmp_files = []
        for i in range(2):
            tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix=f'_multi_{i}.txt', delete=False)
            tmp_file.write(f"Multi attachment content {i}")
            tmp_file.close()
            tmp_files.append(tmp_file.name)

        # Track temp files for cleanup
        self.temp_files.extend(tmp_files)

        # Create mock attachments
        mock_attachments = []
        for tmp_file_path in tmp_files:
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path
            mock_attachments.append(mock_attachment)

        notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test with multiple attachments",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            attachments=mock_attachments,
        )

        assert isinstance(notification, Notification)

        # Verify both attachments were stored
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 2

        # Get attachment names and verify all expected files are there
        attachment_names = [attachment.name for attachment in attachments]
        expected_names = [os.path.basename(tmp_file) for tmp_file in tmp_files]
        assert sorted(attachment_names) == sorted(expected_names)

    def test_persist_one_off_notification_with_multiple_attachments(self):
        """Test persisting one-off notification with multiple attachments"""
        backend = DjangoDbNotificationBackend()

        # Create multiple temporary files for testing
        tmp_files = []
        for i in range(2):
            tmp_file = tempfile.NamedTemporaryFile(mode='w', suffix=f'_oneoff_multi_{i}.txt', delete=False)
            tmp_file.write(f"One-off multi attachment content {i}")
            tmp_file.close()
            tmp_files.append(tmp_file.name)

        # Track temp files for cleanup
        self.temp_files.extend(tmp_files)

        # Create mock attachments
        mock_attachments = []
        for tmp_file_path in tmp_files:
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path
            mock_attachments.append(mock_attachment)

        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="multi@example.com",
            first_name="Jane",
            last_name="Smith",
            notification_type=NotificationTypes.EMAIL.value,
            title="One-off with multiple attachments",
            body_template="test",
            context_name="test",
            context_kwargs={},
            attachments=mock_attachments,
        )

        assert isinstance(one_off_notification, OneOffNotification)

        # Verify both attachments were stored
        notification_db_record = NotificationModel.objects.get(id=one_off_notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 2

        # Get attachment names and verify all expected files are there
        attachment_names = [attachment.name for attachment in attachments]
        expected_names = [os.path.basename(tmp_file) for tmp_file in tmp_files]
        assert sorted(attachment_names) == sorted(expected_names)

    def test_persist_notification_with_empty_attachments_list(self):
        """Test persisting notification with empty attachments list"""
        backend = DjangoDbNotificationBackend()

        notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test without attachments",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            attachments=[],  # Empty list
        )

        assert isinstance(notification, Notification)

        # Verify no attachments were stored
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 0

    def test_persist_notification_with_none_attachments(self):
        """Test persisting notification with None attachments"""
        backend = DjangoDbNotificationBackend()

        notification = backend.persist_notification(
            user_id=self.user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test with None attachments",
            body_template="test",
            context_name="test",
            context_kwargs={},
            send_after=None,
            attachments=None,  # None
        )

        assert isinstance(notification, Notification)

        # Verify no attachments were stored
        notification_db_record = NotificationModel.objects.get(id=notification.id)
        attachments = notification_db_record.attachments.all()
        assert len(attachments) == 0

    def test_store_attachments_with_file_read_error(self):
        """Test _store_attachments method with file that cannot be read"""
        backend = DjangoDbNotificationBackend()

        # Create mock attachment with non-existent file path
        mock_attachment = Mock(spec=NotificationAttachment)
        mock_attachment.file_path = "/non/existent/file/path.txt"

        # Test should raise FileNotFoundError
        with pytest.raises(FileNotFoundError):
            backend._store_attachments([mock_attachment])

    def test_serialize_one_off_notification_with_attachments(self):
        """Test serialization of one-off notification includes attachments"""
        backend = DjangoDbNotificationBackend()

        # Create a temporary file for testing
        with tempfile.NamedTemporaryFile(mode='w', suffix='.txt', delete=False) as tmp_file:
            tmp_file.write("Serialization test content")
            tmp_file_path = tmp_file.name

        # Track temp file for cleanup
        self.temp_files.append(tmp_file_path)

        # Create mock attachment
        mock_attachment = Mock(spec=NotificationAttachment)
        mock_attachment.file_path = tmp_file_path

        one_off_notification = backend.persist_one_off_notification(
            email_or_phone="serialize@example.com",
            first_name="Test",
            last_name="User",
            notification_type=NotificationTypes.EMAIL.value,
            title="Serialization test",
            body_template="test",
            context_name="test",
            context_kwargs={},
            attachments=[mock_attachment],
        )

        # Verify serialization includes attachments
        assert isinstance(one_off_notification, OneOffNotification)
        assert hasattr(one_off_notification, 'attachments')
        assert len(one_off_notification.attachments) == 1

        attachment = one_off_notification.attachments[0]
        assert attachment.filename == os.path.basename(tmp_file_path)
        assert attachment.content_type == 'application/octet-stream'
        assert attachment.size > 0

    def test_serialize_attachment_checksum_calculation(self):
        """Test that _serialize_attachment correctly calculates SHA-256 checksum"""
        backend = DjangoDbNotificationBackend()

        # Create a temporary file with known content
        test_content = b"Test content for checksum calculation"
        expected_checksum = hashlib.sha256(test_content).hexdigest()

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(test_content)
            tmp_file_path = tmp_file.name

        self.temp_files.append(tmp_file_path)

        try:
            # Create mock attachment and store it
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path

            stored_attachments = backend._store_attachments([mock_attachment])
            attachment_instance = stored_attachments[0]

            # Set notification (required for save)
            notification_instance = NotificationModel.objects.create(
                user_id=str(self.user.pk),
                notification_type=NotificationTypes.EMAIL.value,
                title="Checksum test",
                body_template="test",
                context_name="test",
                context_kwargs={},
            )
            attachment_instance.notification = notification_instance
            attachment_instance.save()

            # Test the _serialize_attachment method
            stored_attachment = backend._serialize_attachment(attachment_instance)

            # Verify checksum is calculated correctly
            assert stored_attachment.checksum == expected_checksum
            assert len(stored_attachment.checksum) == 64  # SHA-256 hex length
            assert stored_attachment.filename == os.path.basename(tmp_file_path)
            assert stored_attachment.size == len(test_content)

        finally:
            # Cleanup is handled by tearDown method
            pass

    def test_serialize_attachment_checksum_with_empty_file(self):
        """Test checksum calculation with an empty file"""
        backend = DjangoDbNotificationBackend()

        # Empty content should have specific SHA-256 hash
        test_content = b""
        expected_checksum = hashlib.sha256(test_content).hexdigest()

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            # Don't write anything - file is empty
            tmp_file_path = tmp_file.name

        self.temp_files.append(tmp_file_path)

        try:
            # Create mock attachment and store it
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path

            stored_attachments = backend._store_attachments([mock_attachment])
            attachment_instance = stored_attachments[0]

            # Set notification (required for save)
            notification_instance = NotificationModel.objects.create(
                user_id=str(self.user.pk),
                notification_type=NotificationTypes.EMAIL.value,
                title="Empty checksum test",
                body_template="test",
                context_name="test",
                context_kwargs={},
            )
            attachment_instance.notification = notification_instance
            attachment_instance.save()

            # Test the _serialize_attachment method
            stored_attachment = backend._serialize_attachment(attachment_instance)

            # Verify checksum for empty file
            assert stored_attachment.checksum == expected_checksum
            assert stored_attachment.checksum == "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"  # SHA-256 of empty string

        finally:
            # Cleanup is handled by tearDown method
            pass

    def test_serialize_attachment_checksum_with_binary_content(self):
        """Test checksum calculation with binary content"""
        backend = DjangoDbNotificationBackend()

        # Create binary content
        test_content = bytes(range(256))  # All possible byte values
        expected_checksum = hashlib.sha256(test_content).hexdigest()

        with tempfile.NamedTemporaryFile(delete=False) as tmp_file:
            tmp_file.write(test_content)
            tmp_file_path = tmp_file.name

        self.temp_files.append(tmp_file_path)

        try:
            # Create mock attachment and store it
            mock_attachment = Mock(spec=NotificationAttachment)
            mock_attachment.file_path = tmp_file_path

            stored_attachments = backend._store_attachments([mock_attachment])
            attachment_instance = stored_attachments[0]

            # Set notification (required for save)
            notification_instance = NotificationModel.objects.create(
                user_id=str(self.user.pk),
                notification_type=NotificationTypes.EMAIL.value,
                title="Binary checksum test",
                body_template="test",
                context_name="test",
                context_kwargs={},
            )
            attachment_instance.notification = notification_instance
            attachment_instance.save()

            # Test the _serialize_attachment method
            stored_attachment = backend._serialize_attachment(attachment_instance)

            # Verify checksum for binary content
            assert stored_attachment.checksum == expected_checksum
            assert len(stored_attachment.checksum) == 64  # SHA-256 hex length
            assert stored_attachment.size == len(test_content)

        finally:
            # Cleanup is handled by tearDown method
            pass

    def test_serialize_attachment_checksum_with_no_file(self):
        """Test _serialize_attachment when attachment has no file"""
        backend = DjangoDbNotificationBackend()

        # Create attachment without file
        attachment_instance = AttachmentModel(
            name="no_file_attachment",
            mime_type="text/plain",
            size=0,
        )

        # Don't set the file field

        # Test the _serialize_attachment method
        stored_attachment = backend._serialize_attachment(attachment_instance)

        # Should have empty checksum when no file
        assert stored_attachment.checksum == ""
        assert stored_attachment.filename == "no_file_attachment"
        assert stored_attachment.size == 0

    def test_serialize_attachment_checksum_with_file_read_error(self):
        """Test _serialize_attachment handles file read errors gracefully"""
        backend = DjangoDbNotificationBackend()

        # Create attachment with mock file that raises OSError
        attachment_instance = AttachmentModel(
            name="error_file_attachment",
            mime_type="text/plain",
            size=100,
        )

        # Mock the file to raise an OSError when read
        mock_file = Mock()
        mock_file.seek.side_effect = OSError("File read error")
        attachment_instance.file = mock_file

        # Test the _serialize_attachment method
        stored_attachment = backend._serialize_attachment(attachment_instance)

        # Should have empty checksum when file read fails
        assert stored_attachment.checksum == ""
        assert stored_attachment.filename == "error_file_attachment"
        assert stored_attachment.size == 100



import datetime
import hashlib
import os
import uuid
from collections.abc import Iterable
from typing import cast

from django.core.files.base import ContentFile
from django.db.models import Q, QuerySet

from vintasend.constants import NotificationStatus, NotificationTypes
from vintasend.exceptions import (
    NotificationCancelError,
    NotificationNotFoundError,
    NotificationUpdateError,
    NotificationUserNotFoundError,
)
from vintasend.services.dataclasses import (
    Notification,
    NotificationAttachment,
    OneOffNotification,
    StoredAttachment,
    UpdateNotificationKwargs,
)
from vintasend.services.notification_backends.base import BaseNotificationBackend

from vintasend_django.models import Attachment as AttachmentModel
from vintasend_django.models import Notification as NotificationModel
from vintasend_django.services.attachment_file import DjangoAttachmentFile


class DjangoDbNotificationBackend(BaseNotificationBackend):
    def _get_all_future_notifications_queryset(self) -> QuerySet["NotificationModel"]:
        return NotificationModel.objects.filter(
            Q(send_after__gte=datetime.datetime.now()) | Q(send_after__isnull=False),
            status=NotificationStatus.PENDING_SEND.value,
        ).order_by("created")

    def _get_all_in_app_unread_notifications_queryset(
        self, user_id: int | str | uuid.UUID
    ) -> QuerySet["NotificationModel"]:
        return NotificationModel.objects.filter(
            user_id=str(user_id),
            status=NotificationStatus.SENT.value,
            notification_type=NotificationTypes.IN_APP,
        ).order_by("created")

    def _get_all_pending_notifications_queryset(self) -> QuerySet["NotificationModel"]:
        return NotificationModel.objects.filter(
            Q(send_after__lte=datetime.datetime.now()) | Q(send_after__isnull=True),
            status=NotificationStatus.PENDING_SEND.value,
        ).order_by("created")

    def _paginate_queryset(
        self, queryset: "QuerySet[NotificationModel]", page: int, page_size: int
    ) -> QuerySet["NotificationModel"]:
        return queryset[((page - 1) * page_size) : ((page - 1) * page_size) + page_size]

    def _serialize_user_notification_queryset(
        self, queryset: "QuerySet[NotificationModel]"
    ) -> Iterable[Notification]:
        return (self.serialize_user_notification(n) for n in queryset.iterator())

    def _serialize_notification_queryset(
        self, queryset: "QuerySet[NotificationModel]"
    ) -> Iterable[Notification | OneOffNotification]:
        return (self.serialize_notification(n) for n in queryset.iterator())

    def serialize_notification(self, notification: NotificationModel) -> Notification | OneOffNotification:
        if notification.user_id:
            return self.serialize_user_notification(notification)
        return self.serialize_one_off_notification(notification)

    def serialize_user_notification(self, notification: NotificationModel) -> Notification:
        if not notification.user_id:
            raise NotificationUserNotFoundError("User not found")

        return Notification(
            id=notification.pk,
            user_id=cast(int | str | uuid.UUID, notification.user_id),
            notification_type=notification.notification_type,
            title=notification.title,
            body_template=notification.body_template,
            context_name=notification.context_name,
            context_kwargs=notification.context_kwargs,
            send_after=notification.send_after,
            subject_template=notification.subject_template,
            preheader_template=notification.preheader_template,
            status=notification.status,
        )

    def serialize_one_off_notification(self, notification: NotificationModel) -> OneOffNotification:
        """Serialize Django model to OneOffNotification dataclass"""
        return OneOffNotification(
            id=notification.pk,
            email_or_phone=notification.email_or_phone,
            first_name=notification.first_name,
            last_name=notification.last_name,
            notification_type=notification.notification_type,
            title=notification.title,
            body_template=notification.body_template,
            context_name=notification.context_name,
            context_kwargs=notification.context_kwargs,
            send_after=notification.send_after,
            subject_template=notification.subject_template,
            preheader_template=notification.preheader_template,
            status=notification.status,
            attachments=[self._serialize_attachment(att) for att in notification.attachments.all()],
        )

    def _serialize_attachment(self, attachment) -> StoredAttachment:
        """
        Convert Django attachment model to StoredAttachment.

        Calculates SHA-256 checksum of the file content.
        """
        # Calculate checksum by reading the file content
        checksum = ""
        if attachment.file:
            try:
                # Read file content and calculate SHA-256 hash
                attachment.file.seek(0)  # Ensure we're at the beginning
                file_content = attachment.file.read()
                checksum = hashlib.sha256(file_content).hexdigest()
                attachment.file.seek(0)  # Reset file position
            except OSError:
                # If file reading fails, use empty checksum
                checksum = ""

        return StoredAttachment(
            id=str(attachment.pk),
            filename=attachment.name,
            content_type=attachment.mime_type,
            checksum=checksum,
            size=attachment.size or 0,
            created_at=attachment.created,
            file=DjangoAttachmentFile(attachment),
        )

    def _store_attachments(self, attachments: list[NotificationAttachment]) -> list:
        """Store attachments and return stored attachment objects"""
        stored_attachments = []

        for attachment in attachments:


            # Handle different attachment input types (simplified for now)
            file_content = b''
            file_name = 'attachment'
            mime_type = 'application/octet-stream'

            # Handle different attachment types
            if hasattr(attachment, 'file_path'):
                with open(attachment.file_path, 'rb') as f:
                    file_content = f.read()
                file_name = os.path.basename(attachment.file_path)
            elif hasattr(attachment, 'file_bytes'):
                file_content = attachment.file_bytes
                file_name = getattr(attachment, 'file_name', 'attachment')
            elif hasattr(attachment, 'file_obj'):
                file_obj = attachment.file_obj
                file_obj.seek(0)
                file_content = file_obj.read()
                file_name = getattr(attachment, 'file_name', 'attachment')
            else:
                raise ValueError(
                    f"Unsupported attachment type: {type(attachment)}. "
                    "Attachment must have 'file_path', 'file_bytes', or 'file_obj'."
                )

            # Create attachment record in database
            attachment_instance = AttachmentModel(
                name=file_name,
                mime_type=mime_type,
                size=len(file_content),
            )

            # Save file content to storage
            attachment_instance.file.save(
                file_name,
                ContentFile(file_content),
                save=False  # Don't save the model instance yet
            )

            stored_attachments.append(attachment_instance)

        return stored_attachments

    def persist_notification(
        self,
        user_id: int | str | uuid.UUID,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None,
        subject_template: str | None = None,
        preheader_template: str | None = None,
        adapter_extra_parameters: dict | None = None,
        attachments: list[NotificationAttachment] | None = None,
    ) -> Notification:
        notification_instance = NotificationModel.objects.create(
            user_id=str(user_id),
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template or "",
            preheader_template=preheader_template or "",
            adapter_extra_parameters=adapter_extra_parameters,
        )

        # Store attachments relationship
        if attachments:
            stored_attachments = self._store_attachments(attachments)
            for attachment in stored_attachments:
                attachment.notification = notification_instance
                attachment.save()

        return self.serialize_user_notification(notification_instance)

    def persist_one_off_notification(
        self,
        email_or_phone: str,
        first_name: str,
        last_name: str,
        notification_type: str,
        title: str,
        body_template: str,
        context_name: str,
        context_kwargs: dict[str, uuid.UUID | str | int],
        send_after: datetime.datetime | None = None,
        subject_template: str = "",
        preheader_template: str = "",
        adapter_extra_parameters: dict | None = None,
        attachments: list[NotificationAttachment] | None = None,
    ) -> OneOffNotification:
        """Create and store a one-off notification"""

        notification_instance = NotificationModel.objects.create(
            user=None,  # No user for one-off notifications
            email_or_phone=email_or_phone,
            first_name=first_name,
            last_name=last_name,
            notification_type=notification_type,
            title=title,
            body_template=body_template,
            context_name=context_name,
            context_kwargs=context_kwargs,
            send_after=send_after,
            subject_template=subject_template or "",
            preheader_template=preheader_template or "",
            adapter_extra_parameters=adapter_extra_parameters,
        )

        # Store attachments relationship
        if attachments:
            stored_attachments = self._store_attachments(attachments)
            for attachment in stored_attachments:
                attachment.notification = notification_instance
                attachment.save()

        return self.serialize_one_off_notification(notification_instance)

    def persist_notification_update(
        self, notification_id: int | str | uuid.UUID, updated_data: UpdateNotificationKwargs
    ) -> Notification | OneOffNotification:
        records_updated = NotificationModel.objects.filter(
            id=str(notification_id), status=NotificationStatus.PENDING_SEND.value
        ).update(**updated_data)

        if records_updated == 0:
            raise NotificationUpdateError(
                "Failed to update notification, it may have already been sent"
            )
        return self.serialize_notification(NotificationModel.objects.get(id=str(notification_id)))

    def mark_pending_as_sent(self, notification_id: int | str | uuid.UUID) -> Notification | OneOffNotification:
        records_updated = NotificationModel.objects.filter(
            id=str(notification_id), status=NotificationStatus.PENDING_SEND.value
        ).update(status=NotificationStatus.SENT.value)
        if records_updated == 0:
            raise NotificationUpdateError("Failed to update notification status")
        return self.serialize_notification(NotificationModel.objects.get(id=str(notification_id)))

    def mark_pending_as_failed(self, notification_id: int | str | uuid.UUID) -> Notification | OneOffNotification:
        records_updated = NotificationModel.objects.filter(
            id=str(notification_id), status=NotificationStatus.PENDING_SEND.value
        ).update(status=NotificationStatus.FAILED.value)
        if records_updated == 0:
            raise NotificationUpdateError("Failed to update notification status")
        return self.serialize_notification(NotificationModel.objects.get(id=str(notification_id)))

    def mark_sent_as_read(self, notification_id: int | str | uuid.UUID) -> Notification | OneOffNotification:
        records_updated = NotificationModel.objects.filter(
            id=str(notification_id), status=NotificationStatus.SENT.value
        ).update(status=NotificationStatus.READ.value)
        if records_updated == 0:
            raise NotificationUpdateError("Failed to update notification status")
        return self.serialize_notification(NotificationModel.objects.get(id=str(notification_id)))

    def cancel_notification(self, notification_id: int | str | uuid.UUID) -> None:
        records_updated = NotificationModel.objects.filter(
            id=str(notification_id), status=NotificationStatus.PENDING_SEND.value
        ).update(status=NotificationStatus.CANCELLED.value)

        if records_updated == 0:
            raise NotificationCancelError("Failed to delete notification")

    def get_notification(
        self, notification_id: int | str | uuid.UUID, for_update=False
    ) -> Notification | OneOffNotification:
        """Get notification by ID, supporting both regular and one-off notifications"""
        queryset = NotificationModel.objects.exclude(status=NotificationStatus.CANCELLED.value)

        if for_update:
            queryset = queryset.select_for_update()
        try:
            notification_instance = queryset.get(id=str(notification_id))
        except NotificationModel.DoesNotExist as e:
            raise NotificationNotFoundError("Notification not found") from e

        # Check if it's a one-off notification (no user) or regular notification
        return self.serialize_notification(notification_instance)

    def _get_one_off_notification(self, notification_id: int | str | uuid.UUID) -> OneOffNotification:
        """Retrieve one-off notification from storage"""
        try:
            notification_instance = NotificationModel.objects.exclude(
                status=NotificationStatus.CANCELLED.value
            ).get(id=str(notification_id), user__isnull=True)
        except NotificationModel.DoesNotExist as e:
            raise NotificationNotFoundError(f"One-off notification {notification_id} not found") from e

        return self.serialize_one_off_notification(notification_instance)

    def get_all_pending_notifications(self) -> Iterable[Notification | OneOffNotification]:
        """Return both regular notifications and one-off notifications that are pending"""
        queryset = self._get_all_pending_notifications_queryset()

        # Separate regular notifications (with user) from one-off notifications (without user)
        all_notifications: list[Notification | OneOffNotification] = []

        for notification in queryset:
            if notification.user:
                all_notifications.append(self.serialize_notification(notification))
            else:
                all_notifications.append(self.serialize_one_off_notification(notification))

        return all_notifications

    def get_pending_notifications(self, page: int, page_size: int) -> Iterable[Notification | OneOffNotification]:
        return self._serialize_notification_queryset(
            self._paginate_queryset(
                self._get_all_pending_notifications_queryset(),
                page,
                page_size,
            )
        )

    def filter_all_in_app_unread_notifications(
        self,
        user_id: int | str | uuid.UUID,
    ) -> Iterable[Notification]:
        return self._serialize_user_notification_queryset(
            self._get_all_in_app_unread_notifications_queryset(user_id),
        )

    def filter_in_app_unread_notifications(
        self,
        user_id: int | str | uuid.UUID,
        page: int = 1,
        page_size: int = 10,
    ) -> Iterable[Notification]:
        return self._serialize_user_notification_queryset(
            self._paginate_queryset(
                self._get_all_in_app_unread_notifications_queryset(user_id),
                page,
                page_size,
            )
        )

    def get_all_future_notifications(self) -> Iterable["Notification | OneOffNotification"]:
        return self._serialize_notification_queryset(self._get_all_future_notifications_queryset())

    def get_future_notifications(self, page: int, page_size: int) -> Iterable["Notification | OneOffNotification"]:
        return self._serialize_notification_queryset(
            self._paginate_queryset(self._get_all_future_notifications_queryset(), page, page_size)
        )

    def get_all_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID
    ) -> Iterable["Notification | OneOffNotification"]:
        return self._serialize_user_notification_queryset(
            self._get_all_future_notifications_queryset().filter(user_id=str(user_id))
        )

    def get_future_notifications_from_user(
        self, user_id: int | str | uuid.UUID, page: int, page_size: int
    ) -> Iterable["Notification | OneOffNotification"]:
        return self._serialize_user_notification_queryset(
            self._paginate_queryset(
                self._get_all_future_notifications_queryset().filter(user_id=str(user_id)),
                page,
                page_size,
            )
        )

    def get_user_email_from_notification(self, notification_id: int | str | uuid.UUID) -> str:
        notification_user = (
            NotificationModel.objects.select_related("user").get(id=str(notification_id)).user
        )
        if not notification_user or not notification_user.is_active:
            raise NotificationUserNotFoundError("User not found")
        return notification_user.email

    def store_context_used(
        self,
        notification_id: int | str | uuid.UUID,
        context: dict,
        adapter_import_str: str,
    ) -> None:
        NotificationModel.objects.filter(id=str(notification_id)).update(
            context_used=context, adapter_used=adapter_import_str
        )



import uuid

from django.core import mail

import pytest
from vintasend.constants import NotificationStatus, NotificationTypes
from vintasend.exceptions import (
    NotificationTemplateRenderingError,
)
from vintasend.services.dataclasses import Notification, OneOffNotification
from vintasend.services.notification_backends.stubs.fake_backend import FakeFileBackend

from vintasend_django.services.notification_adapters.django_email import (
    DjangoEmailNotificationAdapter,
)
from vintasend_django.test_helpers import VintaSendDjangoTestCase


class DjangoEmailNotificationAdapterTestCase(VintaSendDjangoTestCase):
    def tearDown(self) -> None:
        mail.outbox = []
        FakeFileBackend(database_file_name="django-email-adapter-test-notifications.json").clear()
        return super().tearDown()

    def create_notification(self, user):
        return Notification(
            id=uuid.uuid4(),
            user_id=user.id,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test Notification",
            body_template="Test Body",
            context_name="test_context",
            context_kwargs={"test": "test"},
            send_after=None,
            subject_template="Test Subject",
            preheader_template="Test Preheader",
            status=NotificationStatus.PENDING_SEND.value,
        )

    def create_notification_context(self):
        return {"foo": "bar"}

    def test_send_notification(self):
        user = self.create_user(email="testadapter@example.com")
        notification = self.create_notification(user)
        context = self.create_notification_context()

        backend = FakeFileBackend(database_file_name="django-email-adapter-test-notifications.json")
        backend.notifications.append(notification)
        backend._store_notifications()

        adapter = DjangoEmailNotificationAdapter(
            "vintasend.services.notification_template_renderers.stubs.fake_templated_email_renderer.FakeTemplateRenderer",
            "vintasend.services.notification_backends.stubs.fake_backend.FakeFileBackend",
            backend_kwargs={"database_file_name": "django-email-adapter-test-notifications.json"},
        )

        adapter.send(notification, context)

        assert len(mail.outbox) == 1
        email = mail.outbox[0]
        assert email.subject == notification.subject_template
        assert email.body == notification.body_template
        assert email.to == ["testemail@example.com"]  # This is the email that the FakeFileBackend returns

    def test_send_notification_with_render_error(self):
        user = self.create_user(email="testadapter@example.com")
        notification = self.create_notification(user)
        context = self.create_notification_context()

        backend = FakeFileBackend(database_file_name="django-email-adapter-test-notifications.json")
        backend.notifications.append(notification)
        backend._store_notifications()

        adapter = DjangoEmailNotificationAdapter(
            "vintasend.services.notification_template_renderers.stubs.fake_templated_email_renderer.FakeTemplateRendererWithException",
            "vintasend.services.notification_backends.stubs.fake_backend.FakeFileBackend",
            backend_kwargs={"database_file_name": "django-email-adapter-test-notifications.json"},
        )
        with pytest.raises(NotificationTemplateRenderingError):
            adapter.send(notification, context)

        assert len(mail.outbox) == 0

    def test_send_one_off_notification(self):
        """Test sending one-off notification"""
        one_off_notification = OneOffNotification(
            id=uuid.uuid4(),
            email_or_phone="oneoff@example.com",
            first_name="John",
            last_name="Doe",
            notification_type=NotificationTypes.EMAIL.value,
            title="One-off Test Notification",
            body_template="Test Body",
            context_name="test_context",
            context_kwargs={"test": "test"},
            send_after=None,
            subject_template="Test Subject",
            preheader_template="Test Preheader",
            status=NotificationStatus.PENDING_SEND.value,
            attachments=[],
        )
        context = self.create_notification_context()

        adapter = DjangoEmailNotificationAdapter(
            "vintasend.services.notification_template_renderers.stubs.fake_templated_email_renderer.FakeTemplateRenderer",
            "vintasend.services.notification_backends.stubs.fake_backend.FakeFileBackend",
            backend_kwargs={"database_file_name": "django-email-adapter-test-one-off.json"},
        )

        adapter.send(one_off_notification, context)

        assert len(mail.outbox) == 1
        email = mail.outbox[0]
        assert email.subject == one_off_notification.subject_template
        assert email.body == one_off_notification.body_template
        assert email.to == ["oneoff@example.com"]

    def create_one_off_notification(self):
        """Helper method to create one-off notification for testing"""
        return OneOffNotification(
            id=uuid.uuid4(),
            email_or_phone="oneoff@example.com",
            first_name="Test",
            last_name="User",
            notification_type=NotificationTypes.EMAIL.value,
            title="Test One-off Notification",
            body_template="Test Body",
            context_name="test_context",
            context_kwargs={"test": "test"},
            send_after=None,
            subject_template="Test Subject",
            preheader_template="Test Preheader",
            status=NotificationStatus.PENDING_SEND.value,
            attachments=[],
        )



from typing import TYPE_CHECKING

from django.template.loader import render_to_string

from vintasend.exceptions import (
    NotificationBodyTemplateRenderingError,
    NotificationPreheaderTemplateRenderingError,
    NotificationSubjectTemplateRenderingError,
)
from vintasend.services.dataclasses import Notification, OneOffNotification
from vintasend.services.notification_template_renderers.base_templated_email_renderer import (
    BaseTemplatedEmailRenderer,
    TemplatedEmail,
)


if TYPE_CHECKING:
    from vintasend.services.notification_service import NotificationContextDict


class DjangoTemplatedEmailRenderer(BaseTemplatedEmailRenderer):
    def render(
        self,
        notification: "Notification | OneOffNotification",
        context: "NotificationContextDict",
        **kwargs
    ) -> TemplatedEmail:
        subject_template = notification.subject_template
        body_template = notification.body_template
        preheader_template = notification.preheader_template

        # Add recipient information to context for one-off notifications
        enhanced_context = context.copy()
        if isinstance(notification, OneOffNotification):
            enhanced_context.update({
                'recipient_email': notification.email_or_phone,
                'recipient_first_name': notification.first_name,
                'recipient_last_name': notification.last_name,
                'recipient_full_name': f"{notification.first_name} {notification.last_name}".strip(),
            })

        try:
            enhanced_context["private_preheader"] = render_to_string(
                preheader_template,
                enhanced_context,
            )
        except Exception as e:  # noqa: BLE001
            raise NotificationPreheaderTemplateRenderingError(
                "Failed to render preheader template"
            ) from e

        try:
            subject = render_to_string(subject_template, enhanced_context)
        except Exception as e:  # noqa: BLE001
            raise NotificationSubjectTemplateRenderingError(
                "Failed to render subject template"
            ) from e

        try:
            body = render_to_string(body_template, enhanced_context)
        except Exception as e:  # noqa: BLE001
            raise NotificationBodyTemplateRenderingError("Failed to render body template") from e

        return TemplatedEmail(subject=subject, body=body)



from django.contrib.auth import get_user_model
from django.db import models
from django.utils.translation import gettext_lazy as _

from model_utils.fields import AutoCreatedField, AutoLastModifiedField

from vintasend_django.constants import NotificationStatusChoices, NotificationTypesChoices


User = get_user_model()

class Notification(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE, null=True, blank=True)
    email_or_phone = models.CharField(max_length=255, blank=True)
    first_name = models.CharField(max_length=255, blank=True)
    last_name = models.CharField(max_length=255, blank=True)
    notification_type = models.CharField(max_length=50, choices=NotificationTypesChoices)
    title = models.CharField(max_length=255)
    status = models.CharField(
        max_length=50, choices=NotificationStatusChoices, default=NotificationStatusChoices.PENDING_SEND
    )
    body_template = models.CharField(max_length=255)

    # Email specific fields
    subject_template = models.CharField(max_length=255, blank=True)
    preheader_template = models.CharField(max_length=255, blank=True)
    context_name = models.CharField(max_length=255, blank=True)
    context_kwargs = models.JSONField(default=dict)

    send_after = models.DateTimeField(null=True)

    created = AutoCreatedField(_("created"), db_index=True)
    modified = AutoLastModifiedField(_("modified"), db_index=True)

    adapter_extra_parameters = models.JSONField(_("extra parameters for the notification adapter"), null=True)

    context_used = models.JSONField(_("context used when notification was sent"), null=True)
    adapter_used = models.CharField(_("adapter used to send the notification"), max_length=255, blank=True)

    objects: models.Manager["Notification"]

    class Meta:
        ordering = ("-created",)

    def __str__(self):
        return f"{self.user} - {self.notification_type} - {self.title} - {self.status}{f' (scheduled to {self.send_after})' if self.send_after else ''}"


class Attachment(models.Model):
    notification = models.ForeignKey(Notification, on_delete=models.CASCADE, related_name="attachments")
    file = models.FileField(upload_to="notifications/attachments/")
    name = models.CharField(max_length=255)
    mime_type = models.CharField(max_length=255, blank=True)
    size = models.PositiveIntegerField(null=True)

    created = AutoCreatedField(_("created"), db_index=True)
    modified = AutoLastModifiedField(_("modified"), db_index=True)

    objects: models.Manager["Attachment"]

    class Meta:
        ordering = ("-created",)

    def __str__(self):
        return f"{self.name} ({self.notification})"



# Generated by Django 5.2.6 on 2025-09-16 18:23

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models
import model_utils.fields


class Migration(migrations.Migration):
    dependencies = [
        ("vintasend_django", "0001_initial"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.AddField(
            model_name="notification",
            name="email_or_phone",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="notification",
            name="first_name",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name="notification",
            name="last_name",
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AlterField(
            model_name="notification",
            name="user",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.CreateModel(
            name="Attachment",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("file", models.FileField(upload_to="notifications/attachments/")),
                ("name", models.CharField(max_length=255)),
                ("mime_type", models.CharField(blank=True, max_length=255)),
                ("size", models.PositiveIntegerField(null=True)),
                (
                    "created",
                    model_utils.fields.AutoCreatedField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="created",
                    ),
                ),
                (
                    "modified",
                    model_utils.fields.AutoLastModifiedField(
                        db_index=True,
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="modified",
                    ),
                ),
                (
                    "notification",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="attachments",
                        to="vintasend_django.notification",
                    ),
                ),
            ],
            options={
                "ordering": ("-created",),
            },
        ),
    ]



from typing import BinaryIO

from vintasend.services.dataclasses import AttachmentFile

from vintasend_django.models import Attachment


class DjangoAttachmentFile(AttachmentFile):
    """Django-specific implementation for file access"""

    def __init__(self, attachment: Attachment):
        self.attachment = attachment

    @property
    def name(self) -> str:
        """Get the attachment name"""
        return self.attachment.name

    @property
    def mime_type(self) -> str:
        """Get the attachment MIME type"""
        return self.attachment.mime_type

    def read(self) -> bytes:
        """Read the entire file content"""
        self.attachment.file.seek(0)
        return self.attachment.file.read()

    def stream(self) -> BinaryIO:
        """
        Return a new file stream for large files.

        Each call to this method opens a new file handle. 
        The caller is responsible for closing the returned stream to prevent resource leaks.
        """
        return self.attachment.file.open('rb')

    def url(self, expires_in: int = 3600) -> str:
        """Generate temporary URL if supported"""
        return self.attachment.file.url

    def delete(self) -> None:
        """Delete from storage"""
        if self.attachment.file:
            self.attachment.file.delete(save=False)
        self.attachment.delete()



from django.test import TestCase
from typing import TYPE_CHECKING

from model_bakery import baker

from django.contrib.auth import get_user_model

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser as DjangoUser


User = get_user_model()


class VintaSendDjangoTestCase(TestCase):
    def setUp(self):
        self._user_password = "123456"
        self.user: DjangoUser = baker.prepare(User, email="user@example.com")
        self.user.set_password(self._user_password)
        self.user.save()

    def create_user(self, **kwargs):
        return baker.make(User, **kwargs)



import uuid
from typing import TYPE_CHECKING 

from vintasend.constants import NotificationStatus, NotificationTypes
from vintasend.services.dataclasses import Notification
from vintasend_django.services.notification_template_renderers.django_templated_email_renderer import (
    DjangoTemplatedEmailRenderer,
)
from vintasend_django.test_helpers import VintaSendDjangoTestCase
from django.contrib.auth import get_user_model


if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser as DjangoUser

User = get_user_model()


class DjangoTemplatedEmailRendererTestCase(VintaSendDjangoTestCase):
    def create_notification(self, user: "DjangoUser") -> Notification:
        return Notification(
            id=uuid.uuid4(),
            user_id=user.pk,
            notification_type=NotificationTypes.EMAIL.value,
            title="Test Notification",
            body_template="vintasend_django/emails/test/test_templated_email_body.html",
            context_name="test_context",
            context_kwargs={},
            send_after=None,
            subject_template="vintasend_django/emails/test/test_templated_email_subject.txt",
            preheader_template="vintasend_django/emails/test/test_templated_email_preheader.html",
            status=NotificationStatus.PENDING_SEND.value,
        )

    def create_notification_context(self, notification: Notification):
        return {
            "test_subject": "this_is_my_test_subject_string",
            "test_preheader": "this_is_my_test_preheader_string",
            "test_body": "this_is_my_test_body_string",
        }

    def test_render(self):
        renderer = DjangoTemplatedEmailRenderer()
        user = self.create_user()
        notification = self.create_notification(user)
        context = self.create_notification_context(notification)
        email = renderer.render(notification, context)
        assert "this_is_my_test_subject_string" in email.subject
        assert "this_is_my_test_preheader_string" in email.body
        assert "this_is_my_test_body_string" in email.body



from typing import TYPE_CHECKING, Generic, TypeVar

from django.contrib.auth import get_user_model
from django.core.mail import EmailMessage

from vintasend.app_settings import NotificationSettings
from vintasend.constants import NotificationTypes
from vintasend.services.dataclasses import Notification, OneOffNotification
from vintasend.services.notification_adapters.base import BaseNotificationAdapter
from vintasend.services.notification_backends.base import BaseNotificationBackend
from vintasend.services.notification_template_renderers.base_templated_email_renderer import (
    BaseTemplatedEmailRenderer,
)


if TYPE_CHECKING:
    from vintasend.services.notification_service import NotificationContextDict


User = get_user_model()


B = TypeVar("B", bound=BaseNotificationBackend)
T = TypeVar("T", bound=BaseTemplatedEmailRenderer)

class DjangoEmailNotificationAdapter(Generic[B, T], BaseNotificationAdapter[B, T]):
    notification_type = NotificationTypes.EMAIL

    def send(
        self,
        notification: "Notification | OneOffNotification",
        context: "NotificationContextDict",
        headers: dict[str, str] | None = None,
    ) -> None:
        """
        Send the notification to the user through email.

        :param notification: The notification to send (regular or one-off).
        :param context: The context to render the notification templates.
        """
        notification_settings = NotificationSettings()

        # Get recipient information based on notification type
        recipient_info = self._get_recipient_info(notification)

        to = [recipient_info["email"]]
        bcc = [email for email in notification_settings.NOTIFICATION_DEFAULT_BCC_EMAILS] or []

        context_with_base_url: "NotificationContextDict" = context.copy()
        context_with_base_url["base_url"] = f"{notification_settings.NOTIFICATION_DEFAULT_BASE_URL_PROTOCOL}://{notification_settings.NOTIFICATION_DEFAULT_BASE_URL_DOMAIN}"

        template = self.template_renderer.render(notification, context_with_base_url)

        email = EmailMessage(
            subject=template.subject.strip(),
            body=template.body,
            from_email=notification_settings.NOTIFICATION_DEFAULT_FROM_EMAIL,
            to=to,
            bcc=bcc,
            headers=headers,
        )
        email.content_subtype = "html"

        # Attach files if any
        self._attach_files(email, notification)

        email.send()

    def _get_recipient_info(self, notification: "Notification | OneOffNotification") -> dict[str, str]:
        """Extract recipient information from notification"""

        if isinstance(notification, OneOffNotification):
            # One-off notification: use provided email/phone and name
            return {
                "email": notification.email_or_phone,
                "full_name": f"{notification.first_name} {notification.last_name}".strip(),
            }
        else:
            # Regular notification: fetch user information
            user_email = self.backend.get_user_email_from_notification(notification.id)
            return {
                "email": user_email,
                "full_name": "",  # Could be extended to get user's full name
            }

    def _attach_files(self, email_message: EmailMessage, notification: "Notification | OneOffNotification") -> None:
        """Attach files to email message"""
        if not hasattr(notification, 'attachments') or not notification.attachments:
            return

        for attachment in notification.attachments:
            try:
                file_data = attachment.file.read()
                # Access Django attachment model for metadata
                django_file = attachment.file
                if hasattr(django_file, 'attachment'):
                    name = django_file.attachment.name
                    mime_type = django_file.attachment.mime_type
                else:
                    # Fallback for basic attachment info
                    name = str(attachment.id) if hasattr(attachment, 'id') else 'attachment'
                    mime_type = 'application/octet-stream'

                email_message.attach(name, file_data, mime_type)
            except Exception as e:
                # Log error but don't break notification sending
                import logging
                logging.warning("Failed to attach file %s: %s", getattr(attachment, 'id', 'unknown'), e)
                continue
